# This file is part of PubMarine.
#
# Foobar is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Foobar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright: 2017, Toshio Kuratomi
# License: LGPLv3+
"""
PubMarine is a simple PubSub framework for Python3's asyncio.

Authors: Toshio Kuratomi <toshio@fedoraproject.org
"""

__version__ = '0.1'
__version_info__ = ('0', '1')

from collections import defaultdict
from weakref import WeakMethod, ref

class PubMarineError(Exception):
    """ Base of all errors specific to PubMarine
    """
    pass


class EventNotFoundError(PubMarineError):
    """ Raised when an event is not handled by this PubPen
    """
    pass


class PubPen:
    """
    A PubPen object coordinates subscription and publication.

    Most programs should create one PubPen instance and then share it between
    all of the objects that wish to communicate with each other.

    """
    def __init__(self, loop, event_list=None):
        """
        :arg loop: Event loop (asyncio compatible) to use.
        :kwarg event_list: If given, event_list is a list of allowed
            event_names.  If not given, any name can be subscribed to on the
            fly.  Dynamic event_lists are convenient.  Statically defined
            lists provide protection against typos.
        """
        self.loop = loop
        if event_list is not None:
            self._event_list = frozenset(event_list)
        else:
            self._event_list = frozenset()

        self._event_handlers = defaultdict(set)

    def subscribe(self, event, callback):
        """ Subscribe a callback to an event

        :arg event: String name of an event to subscribe to
        :callback: The function to call when the event is emitted.  This can
            be any python callable.

        Use :func:`functools.partial` to call the callback with any other
        arguments.

        """
        if self._event_list and event not in self._event_list:
            raise EventNotFoundError('{} is not a registered event' \
                    .format(event))
        try:
            # Add a method
            self._event_handlers[event].add(WeakMethod(callback))
        except TypeError:
            # Add a function
            self._event_handlers[event].add(ref(callback))

    def emit(self, event, *args, **kwargs):
        """ Emit an event

        :arg event: String name of an event to emit

        Other args and keyword args are passed to the callback function.
        """
        if self._event_list and event not in self._event_list:
            raise EventNotFoundError('{} is not a registered event' \
                    .format(event))

        gone = []
        for handler in self._event_handlers[event]:
            # Get the callback from the weakref
            func = handler()
            if func is None:
                # Callback was deleted.  Cleanup the weakref as well
                gone.append(handler)
                continue
            self.loop.call_soon(func, *args, **kwargs)

        # Cleanup any handlers that are no longer around
        for handler in gone:
            self._event_handlers[event].discard(handler)



#!/usr/bin/python3 -tt
#
# Copyright: 2016, Toshio Kuratomi
# License: MIT

import asyncio
from pubmarine import PubPen

class Client:
    def __init__(self, pubpen):
        self.pubpen = pubpen
        self.pubpen.subscribe('server_msg', self.display)

    def display(self, message):
        print(message)

    async def send_message(self):
        for num in range(0, 5):
            print('message: {}'.format(num))
            await asyncio.sleep(2)

class Server:
    def __init__(self, pubpen):
        self.pubpen = pubpen
        self.beats = 0
        self.pubpen.loop.call_later(1, self.heartbeat)

    def heartbeat(self):
        self.pubpen.emit('server_msg', self.beats)
        self.beats += 1
        self.pubpen.loop.call_later(1, self.heartbeat)

if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    pubpen = PubPen(loop)
    server = Server(pubpen)
    client = Client(pubpen)
    loop.run_until_complete(client.send_message())
    loop.close()



#!/usr/bin/python3 -tt
#
# Copyright: 2016, Toshio Kuratomi
# License: MIT

import asyncio
import sys

from pubmarine import PubPen

class Server:
    def __init__(self, pubpen):
        self.pubpen = pubpen
        self.beats = 0
        self.pubpen.loop.call_later(1, self.heartbeat)

        self.pubpen.subscribe('incoming', self.broadcast)

    def broadcast(self, message):
        self.pubpen.emit('outgoing', message)

    def heartbeat(self):
        self.pubpen.emit('outgoing', self.beats)
        self.beats += 1
        self.pubpen.loop.call_later(1, self.heartbeat)


class Client:
    def __init__(self, pubpen, input_queue):
        self.pubpen = pubpen
        self.input = input_queue
        self.pubpen.subscribe('outgoing', self.display)

    def display(self, message):
        print(message)

    @asyncio.coroutine
    def await_input(self):
        while True:
            message = yield from self.input.get()
            if message.strip() == '.':
                self.pubpen.loop.stop()
                break
            self.pubpen.emit('incoming', message)


def get_stdin_data(loop, queue):
    loop.create_task(queue.put(sys.stdin.readline()))


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    queue = asyncio.Queue()
    loop.add_reader(sys.stdin, get_stdin_data, loop, queue)
    pubpen = PubPen(loop)
    server = Server(pubpen)
    client = Client(pubpen, queue)
    loop.run_until_complete(client.await_input())
    loop.close()



from distutils.core import setup

setup(name="pubmarine",
        version="0.2",
        description="PubSub for Python-3.5's asyncio",
        author="Toshio Kuratomi",
        author_email="toshio@fedoraproject.org",
        maintainer="Toshio Kuratomi",
        maintainer_email="toshio@fedoraproject.org",
        url="https://github.com/abadger/pubmarine",
        license="Lesser GNU Public License v3+",
        keywords='pubsub events',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        packages=['pubmarine'],
    )
